/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*global setInterval, window, Image, clearInterval, document*/

/**
 * Application UI canvas module.
 * It is responsible for rendering the canvas content.
 *
 * @module app.ui.canvas
 * @requires {@link app.model.pressure}
 * @namespace app.ui.canvas
 * @memberOf app.ui
 */

// make sure that "app" namespace is created
window.app = window.app || {};

(function defineAppUiCanvas(app) {
    'use strict';

    /**
     * Canvas background image URL.
     *
     * @private
     * @const {string}
     */
    var CANVAS_BACKGROUND_URL = 'images/scale.png',

        /**
         * Animation duration.
         *
         * @private
         * @const {number}
         */
        ANIMATION_DURATION = 2000,

        /**
         * Line height.
         *
         * @private
         * @const {number}
         */
        LINE_HEIGHT = 240,

        /**
         * Default font family.
         *
         * @private
         * @const {string}
         */
        FONT_FAMILY = 'BreezeSans',

        /**
         * UI canvas module reference.
         *
         * @private
         * @type {object}
         */
        uiCanvas = null,

        /**
         * Pressure model module reference.
         *
         * @private
         * @type {object}
         */
        modelPressure = app.model.pressure,

        /**
         * Previous altitude object.
         *
         * @private
         * @type {object}
         */
        previousAltitude = null,

        /**
         * Current altitude object.
         *
         * @private
         * @type {object}
         */
        currentAltitude = null,

        /**
         * Canvas element.
         *
         * @private
         * @type {HTMLElement}
         */
        canvas = null,

        /**
         * Canvas context.
         *
         * @private
         * @type {CanvasRenderingContext2D}
         */
        ctx = null,

        /**
         * Canvas background image.
         *
         * @private
         * @type {Image}
         */
        canvasBackgroundImage = new Image(),

        /**
         * Canvas vertical center position value.
         *
         * @private
         * @param {number}
         */
        verticalCenter = 0,

        /**
         * Canvas horizontal center position value.
         *
         * @private
         * @param {number}
         */
        horizontalCenter = 0,

        /**
         * Check altitude interval id.
         *
         * @private
         * @param {number}
         */
        checkAltitudeInterval = null;

    // create namespace for the module
    app.ui = app.ui || {};
    app.ui.canvas = app.ui.canvas || {};
    uiCanvas = app.ui.canvas;

    /**
     * Clears canvas context.
     *
     * @private
     */
    function clearCanvas() {
        ctx.clearRect(0, 0, canvas.width, canvas.height);
        ctx.textAlign = 'center';
    }

    /**
     * Draws canvas background.
     *
     * @private
     * @param {number} stepShiftForRange
     */
    function drawBackground(stepShiftForRange) {
        if (stepShiftForRange > 0) {
            stepShiftForRange -= LINE_HEIGHT;
        }

        ctx.drawImage(canvasBackgroundImage, 0, stepShiftForRange);
    }

    /**
     * Draws canvas altitude line.
     *
     * @private
     * @param {number} order
     * @param {object} drawLineData
     */
    function drawLine(order, drawLineData) {
        var shift = -LINE_HEIGHT * order + drawLineData.stepShiftForRange,
            fontSizeReducer = Math.abs(shift / LINE_HEIGHT) * 20,
            fontSize = 142 - fontSizeReducer,
            fontVerticalCenterShift = fontSize / 3,
            text = drawLineData.text + order;

        ctx.font = fontSize + 'px ' + FONT_FAMILY;
        ctx.fillText(
            text,
            horizontalCenter,
            verticalCenter + shift + fontVerticalCenterShift
        );
    }

    /**
     * Draws canvas content.
     *
     * @private
     * @param {number} timestamp
     */
    function drawCanvasContent(timestamp) {
        var text = 0,
            timestampDelta = 0,
            timestampFraction = 0,
            altitudeDelta = 0,
            totalShift = 0,
            stepShift = 0,
            stepShiftForRange = 0,
            drawLineData = null,
            animationUp = false,
            i = -3;

        if (!previousAltitude || !timestamp) {
            text = currentAltitude.value;
            stepShiftForRange = 0;
        } else {
            timestampDelta = timestamp - currentAltitude.timestamp;
            timestampFraction = timestampDelta / ANIMATION_DURATION;
            altitudeDelta = currentAltitude.value - previousAltitude.value;
            animationUp = altitudeDelta > 0;
            totalShift = altitudeDelta * LINE_HEIGHT;
            stepShift = totalShift * (1 - ((Math.sin(Math.PI / 2 +
                (timestampFraction * Math.PI)) + 1) / 2));
            text = previousAltitude.value +
                Math[animationUp ? 'floor' : 'ceil'](stepShift / LINE_HEIGHT);
            stepShiftForRange = stepShift % LINE_HEIGHT;
        }

        drawLineData = {
            text: text,
            stepShiftForRange: stepShiftForRange
        };

        clearCanvas();
        drawBackground(stepShiftForRange);

        for (; i <= 3; i += 1) {
            drawLine(i, drawLineData);
        }
    }

    /**
     * Sets interval for checking the current altitude.
     *
     * @private
     */
    function setCheckAltitudeInterval() {
        checkAltitudeInterval = setInterval(
            uiCanvas.checkAltitude, ANIMATION_DURATION
        );
    }

    /**
     * Draws canvas content for current request animation frame.
     *
     * @private
     */
    function drawCanvas() {
        var now = Date.now();

        if (now - currentAltitude.timestamp < ANIMATION_DURATION) {
            drawCanvasContent(now);
            window.webkitRequestAnimationFrame(drawCanvas);
        } else {
            window.webkitRequestAnimationFrame(function onAnimationFrame() {
                drawCanvasContent();
                setCheckAltitudeInterval();
            });
        }
    }

    /**
     * Sets current altitude.
     *
     * @private
     */
    function setCurrentAltitude() {
        currentAltitude = {
            timestamp: Date.now(),
            value: modelPressure.getAltitude()
        };
    }

    /**
     * Checks current altitude.
     *
     * If the current altitude is not equal to the previous one,
     * it starts canvas animation.
     *
     * @memberOf app.ui.canvas
     * @public
     */
    uiCanvas.checkAltitude = function checkAltitude() {
        previousAltitude = currentAltitude;
        setCurrentAltitude();
        if (previousAltitude.value !== currentAltitude.value) {
            clearInterval(checkAltitudeInterval);
            drawCanvas();
        }
    };

    /**
     * Handles load event on canvas background element.
     *
     * Draws canvas content for current request animation frame.
     * Sets interval for checking the current altitude.
     *
     * @private
     */
    function onCanvasBackgroundLoad() {
        drawCanvasContent();
        setCheckAltitudeInterval();
    }

    /**
     * Initializes the UI canvas module.
     *
     * @memberOf app.ui.canvas
     * @public
     */
    uiCanvas.init = function init() {
        canvas = document.getElementById('canvas');
        ctx = canvas.getContext('2d');
        verticalCenter = canvas.height / 2;
        horizontalCenter = canvas.width / 2;

        setCurrentAltitude();
        canvasBackgroundImage.onload = onCanvasBackgroundLoad;
        canvasBackgroundImage.src = CANVAS_BACKGROUND_URL;
    };

})(window.app);
